MINUTE = 60
HOUR = 60 * MINUTE
DAY = 24 * HOUR
WEEK = 7 * DAY

try:
    from tornado.options import options
    DEFAULT_TIMEOUT = options.CACHE_DEFAULT_TIMEOUT
except Exception:
    DEFAULT_TIMEOUT = 2 * DAY   # Domyslny czas na jaki keszujemy

                                
MEMCACHE_NONE = '⪔Noneद⪣'        # Special value to have a distinction between
                                # real None and miss


import logging
from functools import wraps
from constants import DEFAULT_TIMEOUT
from constants import MEMCACHE_NONE
from kalmoize.utils import get
from kalmoize.utils import put


log = logging.getLogger(__name__)

def memoize(key_function, vkey_function=None, timeout=DEFAULT_TIMEOUT,
            allow_none=False, l1=True, use_deepcopy=True, optionaladd=False):
    """
    Decorator is caching result of functions and stores it in memcaches under provided key

      :key_function: functions that transforms args and kwargs to key. No cachonig if it returns None
      :vkey_function: version key function, it allows to invalidate many keys
      :l1: local cache per request
      :use_deepcopy: ---
      :timeout: maximal timeout
      :optionaladd: - store in cache only if it does not exists (for race conditions)
      :MEMOIZE_REFRESH: - enforces storing data into cache
    """
    def decorator(f):
        @wraps(f)
        def _wrapper(*args, **kwargs):
            add_to_cache = optionaladd

            # 1. Creation of keys
            key = key_function
            vkey = vkey_function
            time_out = timeout

            # if MEMOIZE_REFRESH is provided, there is no reading, only forcing value save
            refresh = False
            if kwargs.has_key('MEMOIZE_REFRESH'):
                add_to_cache = False
                refresh = kwargs['MEMOIZE_REFRESH']
                del kwargs['MEMOIZE_REFRESH']

            if callable(key_function):
                key = key_function(*args, **kwargs)

            if key is not None:
                if callable(vkey_function):
                    vkey = vkey_function(*args, **kwargs)

                if callable(timeout):
                    time_out = timeout()

                log.debug("KEY %s", key)

                if not refresh:
                    # 2. Get data and return
                    result = get(key, l1, use_deepcopy=use_deepcopy)
                    if result is not None or result == MEMCACHE_NONE:
                        if result == MEMCACHE_NONE:
                            return
                        log.debug("Z CACHE")
                        return result

            # 3. Generating new data
            result = f(*args, **kwargs)

            if not allow_none and result is None:
                return

            if key is not None:
                put(key, result, time_out, vkey, l1, add_to_cache)

            return result
        return _wrapper
    return decorator


import sys

import logging
import random
from copy import deepcopy
from tornado.options import options
import memcache

from constants import MEMCACHE_NONE

log = logging.getLogger(__name__)

try:   
    PREFIX = options.CACHE_PREFIX
except Exception:
    PREFIX = ""


local_memoize_cache = {}
cache = memcache.Client(options.cache_server_list)

## Funkcje podstawowe
def gen_version():
    """Key version generator matching keys and vkeys
    :return: version number, random int
    :rtype: int
    """
    return random.randint(0, sys.maxsize)

def localcache_put(key, value):
    """Wklada dane do localcache.
    :param key: klucz pod którym zostaną zapamiętane dane
    :type key: str
    :param value: dane do zapamiętania
    """
    global local_memoize_cache
    local_memoize_cache[key] = value

def localcache_put_dict(values_dict):
    """puts dict to local cache.
    :param values_dict: dictionary to update
    :type key: dict
    """
    global local_memoize_cache
    local_memoize_cache.update(values_dict)

def localcache_get(key, use_deepcopy=True):
    """Get data from local cache
    :param key: data key
    :type key: str
    :return: saved data or  MEMCACHE_NONE if None was pushed
    """
    global local_memoize_cache

    if key not in local_memoize_cache:
        return

    if use_deepcopy:
        result = deepcopy(local_memoize_cache[key])
    else:
        result = local_memoize_cache[key]

    if isinstance(result, dict):
        if result.get('vkey') and result.get('version'):
            desired_version = localcache_get(result.get('vkey'), use_deepcopy=use_deepcopy)
            if desired_version != result['version']:
                return
        # return results
        return result

    log.debug("LGET %r=%r", key, result)

    # return results
    return result

def get(key, l1=True, use_deepcopy=True):
    """Gets data from provided key
    :param key: 
    :type key: str
    :param l1: use one request cache or not
    :type l1: bool
    :return: - data stored in cache or  MEMCACHE_NONE if None was put
    """
    global cache
    # get data from local cache
    if l1:
        result = localcache_get(key, use_deepcopy=use_deepcopy)
        if result:
            if isinstance(result, int):
                return
            elif result.get('vkey') and result.get('version'):
                desired_version = _get(result['vkey'], use_deepcopy=use_deepcopy)
                if desired_version != result['version']:
                    return
            return result['value']

    # get data from memcache
    result = cache.get(key)
    log.debug("FROM CACHE: %r", result)
    if not result:
        return

    # checking version
    if result.get('vkey') and result.get('version'):
        desired_version = _get(result['vkey'], False, use_deepcopy=use_deepcopy)
        if desired_version != result['version']:
            return

    # put to local cache and return results
    if l1:
        localcache_put(key, result)
        #local_memoize_cache[key] = result
    log.debug("GET %r=%r", key, result)
    return result['value']

def _put(key, value, timeout=0, l1=True, optionaladd=False):
    """Put data directly to cache.
    :param key: key.
    :type key: str
    :param value: data to put
    :param timeout: timeout in seconds
    :type timeout: int
    :param l1: use l1
    :type l1: bool
    """
    #global local_memoize_cache
    global cache
    if value is None:
        value = MEMCACHE_NONE
    log.debug("_PUT %r=%r", key, value)

    # opcjonalne dodanie elementu do cache tylko wtedy jesli go jeszcze nie ma
    if optionaladd:
        resp = cache.add(key, value, timeout)
    else:
        resp = cache.set(key, value, timeout)

    if l1:
        localcache_put(key, deepcopy(value))
        #local_memoize_cache[key] = deepcopy(value)
    return resp

def _get(key, l1=True, use_deepcopy=True):
    """get data from cache or localcache.
    :param key: key
    :type key: str
    :param l1: check l1 or not
    :type l1: bool
    """
    #global local_memoize_cache
    global cache
    # gt data from localcache
    if l1:
        result = localcache_get(key, use_deepcopy=use_deepcopy)
        if result:
            return result

    # get data from memcache
    result = cache.get(key)
    log.debug("_GET %r=%r", key, result)
    if not result:
        return

    # put to local cache
    if l1:
        localcache_put(key, result)
        #local_memoize_cache[key] = result
    return result


def _get_many(keys, l1=True, use_deepcopy=True):
    """get data from cache or localcache.
    :param keys: list of keys
    :type key: str
    :param l1: use 1 request cache or not
    :type l1: bool
    """
    results = {}
    global cache
    # get data from local cache
    if l1:
        for key in keys:
            result = localcache_get(key, use_deepcopy=use_deepcopy)
            if result:
                results[key] = result

    # get data from memcache
    cache_results = cache.get_multi([k for k in keys if k not in results])
    if cache_results:
        # put to local cache and return results
        if l1:
            localcache_put_dict(cache_results)

        results.update(cache_results)

    return results


def put(key, value, timeout=0, vkey=None, l1=True, optionaladd=False):
    """Put data to cache
    :param key: key
    :type key: str
    :param value: data to put
    :param l1: whether to use one request cache or not
    :type l1: bool
    :param timeout: timeout in seconds
    :type timeout: int
    :param optionaladd: set is optional, for race conditions
    """
    #global local_memoize_cache
    global cache
    log.debug("PUT %r=%r", key,value)
    value_dict = dict(value=None, vkey=None, version=None)
    value_dict['vkey'] = vkey

    if value is None:
        value = MEMCACHE_NONE

    if vkey:
        # creating vkey version
        current_version = _get(vkey, l1)
        if not current_version:
            current_version = gen_version()
            _put(vkey, current_version, timeout, l1, optionaladd)
        value_dict['version'] = current_version

    value_dict['value'] = value
    log.debug("PUT %r = dict: %r", key, value_dict)

    # adding to cache
    if optionaladd:
        resp = cache.add(key, value_dict, timeout)
    else:
        resp = cache.set(key, value_dict, timeout)

    if l1:
        localcache_put(key, deepcopy(value_dict))
        #local_memoize_cache[key] = deepcopy(value_dict)
    return resp


def get_many(keys, l1=True, use_deepcopy=True):
    """Get data fom given keys from l1 and memcache.
    :param keys: keys
    :type key: str list
    :param l1: use l1 cache)
    :type l1: bool
    :return: - returns data or MEMOIZE_NONE when it is None
    """
    results = {}
    l1_vkeys = set()
    l1_vkeys_results = {}

    # get data from l1
    if l1:
        l1_results = {}

        for key in keys:
            result = localcache_get(key, use_deepcopy=use_deepcopy)
            if result:
                # We do always put dict in put method
                # we ensure whether it is not int
                if isinstance(result, int):
                    continue

                l1_results[key] = result

        # getting keys
        l1_vkeys = set([v.get('vkey') for v in l1_results.itervalues() if v.get('vkey') and v.get('version')])
        l1_vkeys_results = _get_many(l1_vkeys, use_deepcopy=use_deepcopy)

        # checking version for local data
        for (k, v) in l1_results.iteritems():
            if v.get('version') and v.get('vkey'):
                if v.get('version') != l1_vkeys_results.get(v.get('vkey')):
                    continue
            results[k] = v

    # get data from memcache, that are not available in l1 cache
    cache_results = cache.get_multi([k for k in keys if k not in results])
    if cache_results:
        # getting version keys
        cache_vkeys = set([v.get('vkey') for v in cache_results.itervalues() if v.get('vkey') and
                                                                                v.get('version') and
                                                                                v.get('vkey') not in l1_vkeys])
        cache_vkeys_results = _get_many(cache_vkeys, False, use_deepcopy=use_deepcopy)
        cache_vkeys_results.update(l1_vkeys_results)

        # checking whether version is correct
        valid_cache_results = {}
        for (k, v) in cache_results.iteritems():
            if v.get('version') and v.get('vkey'):
                if v.get('version') != cache_vkeys_results.get(v.get('vkey')):
                    continue
            valid_cache_results[k] = v

        if l1:
            localcache_put_dict(valid_cache_results)

        results.update(valid_cache_results)

    # delete version data from results
    final_dict = {}
    for (k, v) in results.iteritems():
        final_dict[k] = v['value'] if v is not None else None

    return final_dict


def delete(*keys):
    """invalidate many keys
    :param keys: list of keys
    """
    return invalidate(*keys)

def invalidatev(*keys, **kwargs):
    """Invalidate version keys and set new version
    """
    global local_memoize_cache
    resps = []
    for key in keys:
        v = gen_version()
        resps.append( cache.set(key, v) )
        localcache_put(key, v)
        local_memoize_cache[key] = v
    return resps

def invalidate(*keys, **kwargs):
    """invalidate data.
    :param keys: list of keys
    """
    global local_memoize_cache

    resps = []
    for key in keys:
        resps.append( cache.delete(key) )
        if key in local_memoize_cache:
            del local_memoize_cache[key]
    return resps

def localcache_clear():
    """clears local cache, run it after first request"""
    global local_memoize_cache
    local_memoize_cache = {}


#!/usr/bin/env python


from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
from kalmoize import __version__
setup(name='kalmoize',
    version=__version__,
    packages=['kalmoize'],
    description='Rock solid memoize for Python/Tornado',
    author='Piotr Czesław Izajasz Kalmus-Tucholski',
    author_email='pckalmus@gmail.com',
    url='https://github.com/pekoslaw/kalmoize',
    keywords = ['memcache', 'cache', 'tornado'],
    classifiers = [
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Natural Language :: English",
        "Natural Language :: Polish",
        #"Framework :: Tornado",
        "Environment :: Web Environment",
        "Environment :: Other Environment",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Other/Nonlisted Topic",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        
    ],
    cmdclass = {'build_ext': build_ext},
    ext_modules = [Extension("kalmoize.utils", ["kalmoize/utils.py"]),
                   Extension("kalmoize.decorators", ["kalmoize/decorators.py"]),
                   ]
)